from typing import NamedTuple, Optional, Any, Dict

from bson import ObjectId


class BaseLog:
    def __init__(self):
        self.items = []

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, index):
        return self.items[index]

    def __contains__(self, item):
        return item in self.items

    def __len__(self):
        return len(self.items)

    def __bool__(self):
        return bool(self.items)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.items))

    def append(self, log_item):
        self.items.append(log_item)

    def clear(self):
        self.items.clear()


class Save(NamedTuple):
    op: str = 'save'
    id: Optional[ObjectId] = None


class Insert(NamedTuple):
    op: str = 'insert'
    id: Optional[ObjectId] = None


class UpdateOne(NamedTuple):
    update_data: Dict[str, Any]
    op: str = 'update_one'


class DeleteOne(NamedTuple):
    op: str = 'delete_one'


class Reload(NamedTuple):
    op: str = 'reload'


class SetField(NamedTuple):
    name: str
    value: Any
    op: str = 'set_field'


class ChangeChild(NamedTuple):
    path: str
    name: str
    log_item: NamedTuple
    op: str = 'change_chind'



from yadm.log_items import BaseLog, ChangeChild


class ItemLog(BaseLog):
    def __init__(self, document_item):
        super().__init__()
        self.document_item = document_item

    def append(self, log_item):
        self.items.append(log_item)

        root = self.document_item.__document__
        if root is not None:
            self.document_item.__document__.__log__.append(
                ChangeChild(
                    path=self.document_item.__field_name__,
                    name=self.document_item.__name__,
                    log_item=log_item,
                ),
            )


class DocumentItemMixin:
    """ Mixin for custom all fields values, such as EmbeddedDocument,
        yadm.fields.containers.Container.
    """
    __parent__ = None
    __name__ = None
    __qs__ = None
    __log__ = None

    def __init__(self, *args, **kwargs):
        self.__log__ = ItemLog(self)
        super().__init__(*args, **kwargs)

    @property
    def __document__(self):
        """ Root document.

        .. code-block:: python

                assert doc.f.l[0].__document__ is doc
        """
        obj = self

        while getattr(obj, '__parent__', None) is not None:
            obj = obj.__parent__

        if obj is not self:
            return obj
        else:
            return None

    @property
    def __db__(self):
        """ Database object.

        .. code-block:: python

            assert doc.f.l[0].__db__ is doc.__db__
        """
        document = self.__document__
        if document is not None:
            return document.__db__
        else:
            return None

    @property
    def __qs__(self):
        """ Queryset object.
        """
        document = self.__document__
        if document is not None:
            return document.__qs__
        else:
            return None

    @property
    def __path__(self):
        """ Path to root generator.

        .. code-block:: python

            assert list(doc.f.l[0].__path__) == [doc.f.l[0], doc.f.l, doc.f]
        """
        obj = self

        while getattr(obj, '__parent__', None) is not None:
            yield obj
            obj = obj.__parent__

    @property
    def __path_names__(self):
        """ Path to root generator.

        .. code-block:: python

            assert list(doc.f.l[0].__path__) == [0, 'l', 'f']
        """
        for item in self.__path__:
            yield item.__name__

    @property
    def __field_name__(self):
        """ Dotted field name for MongoDB opperations, like as $set, $push and other...

        .. code-block:: python

            assert doc.f.l[0].__field_name__ == 'f.l.0'
        """
        return '.'.join(reversed([str(i) for i in self.__path_names__]))

    def __get_value__(self, document):
        """ Get value from document with path to self.
        """
        obj = document

        for name in reversed(list(self.__path_names__)):
            if isinstance(name, int):
                obj = obj[name]
            else:
                obj = getattr(obj, name)

        return obj



from bson import ObjectId

from yadm.queryset import BaseQuerySet, NotFoundBehavior, NotFoundError
from yadm.results import UpdateResult, RemoveResult


class AioQuerySet(BaseQuerySet):
    async def __aiter__(self):
        async for raw in self._cursor:
            yield self._from_mongo_one(raw)

    async def _get_one(self, index):
        cursor = self._cursor.skip(index).limit(1)
        try:
            raw = await cursor.__anext__()
        except StopAsyncIteration:
            raise IndexError(index)
        finally:
            await cursor.close()

        return self._from_mongo_one(raw)

    async def find_one(self, criteria=None, projection=None, *, exc=None):
        if isinstance(criteria, ObjectId):
            criteria = {'_id': criteria}

        qs = self.find(criteria=criteria, projection=projection)
        data = await self._collection.find_one(qs._criteria, qs._projection)

        if data is None:
            if exc is not None:
                raise exc(criteria)
            else:
                return None

        return self._from_mongo_one(data, projection=qs._projection)

    async def update(self, update, *, multi=True, upsert=False):
        raw_result = await self._collection.update(
            self._criteria,
            update,
            multi=multi,
            upsert=upsert,
        )
        return UpdateResult(raw_result)

    async def find_and_modify(
            self, update=None, *, upsert=False,
            full_response=False, new=False,
            **kwargs):
        result = await self._collection.find_and_modify(
            query=self._criteria,
            update=update,
            upsert=upsert,
            sort=self._sort or [],
            full_response=full_response,
            new=new,
            **kwargs
        )
        if not full_response:
            return self._from_mongo_one(result)
        else:  # pragma: no cover
            result['value'] = self._from_mongo_one(result['value'])
            return result

    async def remove(self, *, multi=True):
        raw_result = await self._collection.remove(self._criteria, multi=multi)
        return RemoveResult(raw_result)

    async def distinct(self, field):
        return await self._cursor.distinct(field)

    async def count(self):
        return await self._cursor.count()

    async def ids(self):
        async for raw in self.copy(projection={'_id': True})._cursor:
            yield raw['_id']

    async def bulk(self):
        qs = self.copy()
        qs._sort = None
        return {obj.id: obj async for obj in qs}

    async def join(self, *field_names):  # pragma: no cover
        raise NotImplementedError

    async def find_in(self, comparable, field='_id', *,
                      not_found=NotFoundBehavior.SKIP):
        not_found = NotFoundBehavior(not_found)
        hash_docs = {}

        async for doc in self.find({field: {'$in': comparable}}):
            key = getattr(doc, field)
            if key not in hash_docs:
                hash_docs[key] = doc

        for cmp_item in comparable:
            value = hash_docs.get(cmp_item)

            if not_found is NotFoundBehavior.NONE:
                yield value

            elif not_found is NotFoundBehavior.SKIP:
                if value is not None:
                    yield value

            elif not_found is NotFoundBehavior.ERROR:
                if value is not None:
                    yield value
                else:
                    raise NotFoundError("Could not find a document with"
                                        " the field '{}' equal '{}'"
                                        "".format(field, cmp_item))



import pymongo

import yadm.abc as abc
from yadm.database import BaseDatabase
from yadm.serialize import to_mongo, from_mongo

from .queryset import AioQuerySet
from .aggregation import AioAggregator

PYMONGO_VERSION = pymongo.version_tuple

RPS = pymongo.read_preferences


@abc.Database.register
class AioDatabase(BaseDatabase):
    aio = True

    async def insert(self, document, **collection_params):
        document.__db__ = self
        collection = self._get_collection(document.__class__, collection_params)
        document._id = await collection.insert(to_mongo(document))
        document.__changed_clear__()
        return document

    async def save(self, document, **collection_params):
        document.__db__ = self
        collection = self._get_collection(document.__class__, collection_params)
        document._id = await collection.save(to_mongo(document))
        document.__changed_clear__()
        return document

    async def update_one(self, document, reload=True, *,
                         set=None, unset=None, inc=None,
                         push=None, pull=None,
                         **collection_params):
        update_data = {}

        if set:
            update_data['$set'] = set

        if unset:
            if isinstance(unset, dict):
                update_data['$unset'] = unset
            else:
                update_data['$unset'] = {f: True for f in unset}

        if inc:
            update_data['$inc'] = inc

        if push:
            update_data['$push'] = push

        if pull:
            update_data['$pull'] = pull

        if update_data:
            await self._get_collection(document, collection_params).update(
                {'_id': document.id},
                update_data,
                upsert=False,
                multi=False,
            )

        if reload:
            await self.reload(document)

    async def remove(self, document, **collection_params):
        collection = self._get_collection(document.__class__, collection_params)
        return await collection.remove({'_id': document._id})

    async def reload(self, document, new_instance=False, *,
                     projection=None,
                     read_preference=RPS.PrimaryPreferred(),
                     **collection_params):
        collection_params['read_preference'] = read_preference

        qs = self.get_queryset(document.__class__,
                               projection=projection,
                               **collection_params)

        new = await qs.find_one(document.id)

        if new_instance:
            return new
        else:
            document.__raw__.clear()
            document.__raw__.update(new.__raw__)
            document.__cache__.clear()
            document.__changed__.clear()
            return document

    async def get_document(self, document_class, _id, *,
                           exc=None,
                           read_preference=RPS.PrimaryPreferred(),
                           **collection_params):
        collection_params['read_preference'] = read_preference
        col = self.db.get_collection(document_class.__collection__,
                                     **collection_params)

        raw = await col.find_one({'_id': _id})

        if raw:
            doc = from_mongo(document_class, raw)
            doc.__db__ = self
            return doc

        elif exc is not None:
            raise exc((document_class, _id, collection_params))

        else:
            return None

    def get_queryset(self, document_class, *,
                     projection=None,
                     cache=None,
                     **collection_params):
        if projection is None:
            projection = document_class.__default_projection__

        return AioQuerySet(self, document_class,
                           projection=projection,
                           cache=cache,
                           collection_params=collection_params)

    def aggregate(self, document_class, *, pipeline=None, **collection_params):
        return AioAggregator(self, document_class,
                             pipeline=pipeline,
                             collection_params=collection_params)

    def bulk(self, document_class,
             ordered=False, raise_on_errors=True, **collection_params):
        raise NotImplementedError
        # return AioBulk(self, document_class, ordered, raise_on_errors,
        #                collection_params=collection_params)
